"""
An experimental plugin for attaching notes to the synoptic.
Not finished!
"""

import csv
import json
import webbrowser

import requests


elog_url = "http://localhost:8081/femtomaxsynoptic"
"""
ELOG Notes:

All posts after 2016-01-12 14:31:39:
  search http://localhost:8081/femtomaxsynoptic/?mode=CSV1&reverse=0&npp=20&ma=1&da=12&ya=2016&ha=14&na=31&ca=39

"""


def open_note(widget, args):
    mid = args
    webbrowser.open(elog_url + "/" + mid)


def new_note(widget, args):
    code, position = args.split(",")
    webbrowser.open(elog_url + "/?cmd=new&pPosition=%s&pIdentifier=%s" % (code, position))


def load_notes(widget, args):
    return get_notes(widget)


def get_notes(widget):
    result = requests.get(elog_url + "?mode=CSV1")
    lines = list(csv.reader(result.text.encode("utf-8").splitlines()))
    notes = []
    headers = lines[0]
    for line in lines[1:]:
        note = dict(item for item in zip(headers, line))
        position = note.get("Position")
        if position:
            x, y = [float(part) if "." in part else int(part)
                    for part in position.split(";")]
            note["position"] = {"x": x, "y": y}
        notes.append(note)
    widget.js.evaluate("synoptic.setNotes('%s')" % json.dumps(notes))



from threading import Event, Lock
from time import sleep

import tango
from PyQt5 import QtCore
from taurus import Attribute
from taurus.core import TaurusException
from taurus.core.evaluation.evalvalidator import EvaluationAttributeNameValidator
from taurus.core.tango.tangovalidator import (TangoAttributeNameValidator,
                                              TangoDeviceNameValidator)

# We can't use pytango's CaselessDict since it does not keep the original
# case of the keys :(
from .caseless import CaselessDictionary as CaselessDict
from .ttldict import TTLDict


class Registry(QtCore.QThread):

    """
    A thread that handles attribute subscriptions.

    The main purpose of this thread is to offload the setting up and
    tearing down of subscriptions from the main UI thread, to make the
    operation "asynchronous". We don't want the UI to lock up briefly
    every time we do this (which may be very often in the case of a
    large, zoomable synoptic).
    """

    def __init__(self, event_callback, unsubscribe_callback, period=0.5):
        QtCore.QThread.__init__(self)
        self.listeners = CaselessDict()
        self.inverse_listeners = {}
        self.event_callback = event_callback
        self.unsubscribe_callback = unsubscribe_callback
        self.period = period
        self._attributes = None
        self._taurus_attributes = CaselessDict()
        self._config = CaselessDict()
        self._last_event = TTLDict(default_ttl=10)
        self.attribute_validator = TangoAttributeNameValidator()
        self.device_validator = TangoDeviceNameValidator()
        self.eval_validator = EvaluationAttributeNameValidator()
        self.lock = Lock()
        self.stopped = Event()

    def run(self):
        """
        A simple loop checking for changes to the attribute list
        """
        # "batching" the updates should be more efficient than
        # reacting immediately, especially when listeners can come and
        # go quite frequently.
        while not self.stopped.is_set():
            sleep(self.period)
            if self._attributes is not None:
                attributes, self._attributes = self._attributes, None
                with self.lock:
                    self._update(attributes)

    def subscribe(self, models=[]):
        """
        Set the currently subscribed list of models.
        """
        attrs = CaselessDict()
        taurusattrs = self._taurus_attributes
        for model in models:
            if self.device_validator.isValid(model):
                # for convenience, we subscribe to State for any devices
                modelstate = model + "/State"
                attrs[modelstate] = True
                if modelstate not in taurusattrs.keys():
                    try:
                        taurusattrs[modelstate] = Attribute(modelstate)
                    except TaurusException as e:
                        print("Failed to create Taurus Attribute for model %s! %s" % (model, e))
            elif (self.attribute_validator.isValid(model)
                  or self.eval_validator.isValid(model)):
                attrs[model] = True
                if model not in taurusattrs.keys():
                    try:
                        taurusattrs[model] = Attribute(model)
                    except TaurusException as e:
                        print("Failed to create Taurus Attribute for model %s! %s" % (model, e))
                    except Exception as e:
                        print("Failed to create Taurus Attribute for model %s! %s" % (model, e))
            else:
                print("Invalid Taurus model %s!?" % model)
        self._attributes = attrs
        self._taurus_attributes = taurusattrs

    def get_value(self, model):
        evt = self._last_event.get(model)
        if evt:
            return evt.attr_value

    def get_config(self, model):
        return self._config.get(model)

    def handle_event(self, evt_src, *args):
        # lookup the model(s) for this listener and notify them
        models = self.inverse_listeners.get(evt_src, [])
        for model in models:
            self.event_callback(model, evt_src, *args)

    def _update(self, attributes=CaselessDict()):
        """
        Update the subscriptions; add new ones, remove old ones
        """

        listeners = set(k.lower() for k in self.listeners.keys())
        new_attrs = set(attributes) - set(listeners)
        old_attrs = set(listeners) - set(attributes)

        for attr in old_attrs:
            if self._attributes:
                # meaning we got a new list of subscriptions, so
                # no point in continuing with this one.
                return
            self._remove_listener(attr)
            self._last_event.pop(attr, None)

        for attr in new_attrs:
            if self._attributes:
                return
            try:
                self._add_listener(attr)
            except (TypeError, tango.DevFailed) as e:
                print("Failed to setup listener for", attr, e)

        self.unsubscribe_callback(old_attrs)

    def _add_listener(self, model):
        try:
            listener = self.listeners[model] = self._taurus_attributes[model]
            # to make loopkups more efficient, we also keep an "inverted"
            # mapping of listeners->models. But since some models may map
            # to the *same* listener (e.g. eval expressions), this must
            # be a one-to-many map.
            if listener in self.inverse_listeners:
                self.inverse_listeners[listener][model] = True
            else:
                self.inverse_listeners[listener] = CaselessDict([(model, True)])
            listener.addListener(self.handle_event)
            return listener
        except (TaurusException, AttributeError) as e:
            print("Failed to subscribe to model %s! %s" % (model, e))
        except Exception:
            print("Failed to subscribe to model %s!" % model)

    def _remove_listener(self, model):
        listener = self.listeners.pop(model)
        models = self.inverse_listeners[listener]
        models.pop(model)
        if not models:
            self.inverse_listeners.pop(listener)
        listener.removeListener(self.handle_event)

    def get_listener(self, model):
        """
        return the listener for a given model
        """
        if model in self.listeners:
            return self.listeners[model]
        for attr in self.listeners.values():
            if attr.getNormalName().lower() == model.lower():
                return attr

    def stop(self):
        self.stopped.set()

    def clear(self):
        self._attributes.clear()
        self._last_event.clear()
        self._update()



import sys

from taurus import Attribute, Manager
from taurus.core.taurusbasetypes import TaurusEventType, DataFormat
import tango

Manager().changeDefaultPollingPeriod(1000)


def error_str(err):
    if isinstance(err, tango.DevFailed):
        err = err[0]
        return "error"
    return str(err)


# Based on code from the taurus-web project
class TaurusAttribute(object):
    """
    This object is a listener for the taurus attribute value.
    When a attribute changes it sends an event. The event
    triggers a call to *eventReceived*. *eventReceived* will transform
    the change event into a JSON encoded string and sends this
    string through the web socket to the client
    """

    def __init__(self, name, callback):
        self.name = name
        self.callback = callback
        self._last_time = 0
        self.last_value_event = None
        self.last_config_event = None
        self.attribute = Attribute(self.name)
        self.attribute.addListener(self)

    def eventReceived(self, evt_src, evt_type, evt_value):
        """
        Transforms the event into a JSON encoded string and sends this
        string into the web socket
        """

        modelObj = evt_src
        data = {}
        if evt_type == TaurusEventType.Error:
            data["event_type"] = "error"
            data["error"] = error_str(evt_value)
        else:
            if evt_type == TaurusEventType.Config:
                modelObj = evt_src.getParentObj()
                data['event_type'] = "config"
                data['description'] = evt_src.description
                data['label'] = evt_src.label
                data["unit"] = evt_src.unit if evt_src.unit != "No unit" else ""
                data["format"] = evt_src.format
                self.last_config_event = data
            else:
                self._last_time = evt_value.time.tv_sec
                data["event_type"] = "value"
                value = evt_value.value
                quality = evt_value.quality
                fmt = evt_value.data_format
                if fmt == DataFormat._0D:
                    html = modelObj.displayValue(value)
                    if isinstance(value, tango._tango.DevState):
                        value = value.name
                    if isinstance(value, str):
                        html = value.replace("\n", "<br>")
                elif fmt in (DataFormat._1D, DataFormat._2D):
                    # bad, very bad performance! Don't worry for now
                    value = value.tolist()
                    html = "[...]"
                data['value'] = value
                data['html'] = html
                data['quality'] = str(quality)
                data['time'] = evt_value.time.tv_sec
                data['type'] = str(evt_value.type)  # can this change..?
                self.last_value_event = data

        data['model'] = modelObj.getNormalName()
        self.write_message(data)

    def write_message(self, message):
        self.callback(message)

    def clear(self):
        try:
            self.attribute.removeListener(self)
            del self.attribute

        except tango.DevFailed as e:
            print("Could not unsubscribe %s: %r" % (self.name, e), file=sys.stderr)

    # def __del__(self):
    #     print "__del__ listener"



"""A caseless dictionary implementation."""

from collections.abc import MutableMapping


class CaselessDictionary(MutableMapping):

    """
    A dictionary-like object which ignores but preserves the case of strings.

    Example::

        >>> cdict = CaselessDictionary()

    Access is case-insensitive::

        >>> cdict['a'] = 1
        >>> cdict['A']
        1

    As is writing::

        >>> cdict['key'] = 123
        >>> cdict['KeY'] = 456
        >>> cdict['key']
        456

    And deletion::

        >>> del cdict['A']
        >>> 'a' in cdict
        False
        >>> 'A' in cdict
        False

    However, the case of keys is preserved (the case of overridden keys will be
    the first one which was set)::

        >>> cdict['aBcDeF'] = 1
        >>> sorted(list(cdict))
        ['aBcDeF', 'key']
    """

    def __init__(self, *args, **kwargs):
        self.__dict__["_dict"] = {}
        temp_dict = dict(*args, **kwargs)
        for key, value in temp_dict.items():
            if isinstance(key, str):
                key = CaselessString.make_caseless(key)
            self._dict[key] = value

    def __getitem__(self, key):
        return self._dict[CaselessString.make_caseless(key)]

    def __setitem__(self, key, value):
        self._dict[CaselessString.make_caseless(key)] = value

    def __delitem__(self, key):
        del self._dict[CaselessString.make_caseless(key)]

    def __contains__(self, key):
        return CaselessString.make_caseless(key) in self._dict

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def keys(self):
        # convert back to ordinary strings
        return [str(k) for k in self._dict]

    def items(self):
        return zip(self.keys(), self.values())


class CaselessString(object):

    """
    A mixin to make a string subclass case-insensitive in dict lookups.
    """

    def __hash__(self):
        return hash(self.lower())

    def __eq__(self, other):
        return self.lower() == other.lower()

    def __cmp__(self, other):
        return self.lower().__cmp__(other.lower())

    @classmethod
    def make_caseless(cls, string):
        return CaselessStr(string)


class CaselessStr(CaselessString, str):
    pass



"""
TTL dictionary

Tricks / features:
 - calling len() will remove expired keys
 - __repr__() might show expired values, doesn't remove expired ones
"""

__all__ = ['TTLDict']
__version__ = '0.0.4'

import time
from collections.abc import MutableMapping
from threading import RLock


class TTLDict(MutableMapping):
    """
    Dictionary with TTL
    Extra args and kwargs are passed to initial .update() call
    """
    def __init__(self, default_ttl, *args, **kwargs):
        self._default_ttl = default_ttl
        self._values = {}
        self._lock = RLock()
        self.update(*args, **kwargs)

    def __repr__(self):
        return '<TTLDict@%#08x; ttl=%r, v=%r;>' % (id(self), self._default_ttl, self._values)

    def set_ttl(self, key, ttl, now=None):
        """
        Set TTL for the given key
        """
        if now is None:
            now = time.time()
        with self._lock:
            _expire, value = self._values[key]
            self._values[key] = (now + ttl, value)

    def get_ttl(self, key, now=None):
        """
        Return remaining TTL for a key
        """
        if now is None:
            now = time.time()
        with self._lock:
            expire, _value = self._values[key]
            return expire - now

    def expire_at(self, key, timestamp):
        """
        Set the key expire timestamp
        """
        with self._lock:
            _expire, value = self._values[key]
            self._values[key] = (timestamp, value)

    def is_expired(self, key, now=None, remove=False):
        """
        Check if key has expired
        """
        with self._lock:
            if now is None:
                now = time.time()
            expire, _value = self._values[key]
            if expire is None:
                return False
            expired = expire < now
            if expired and remove:
                self.__delitem__(key)
            return expired

    def __len__(self):
        with self._lock:
            for key in self._values.keys():
                self.is_expired(key, remove=True)
            return len(self._values)

    def __iter__(self):
        with self._lock:
            for key in self._values.keys():
                if not self.is_expired(key, remove=True):
                    yield key

    def __setitem__(self, key, value):
        with self._lock:
            if self._default_ttl is None:
                expire = None
            else:
                expire = time.time() + self._default_ttl
            self._values[key] = (expire, value)

    def __delitem__(self, key):
        with self._lock:
            del self._values[key]

    def __getitem__(self, key):
        with self._lock:
            self.is_expired(key, remove=True)
            return self._values[key][1]

    def __contains__(self, key):
        try:
            _ = TTLDict.__getitem__(self, key)
            return True
        except KeyError:
            return False



#!/usr/bin/env python

from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="svgsynoptic2",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="Widget for displaying a SVG synoptic.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Johan Forsberg",
    author_email="johan.forsberg@maxiv.lu.se",
    license="GPLv3",
    python_requires=">=3.6",
    install_requires=["pytango>=9.2.1", "taurus>=4.5", "pyqtwebengine"],
    url="https://gitlab.com/MaxIV/lib-maxiv-svgsynoptic",
    packages=find_packages(),
    include_package_data=True,
    package_data={"svgsynoptic2": ["web/js/*.js", "web/js/libs/*.js", "web/css/*.css"]},
)
